import numpy as np
import numpy.matlib as ml

import topology.util.math as mutil

def through(vertices, indices):
    """ Calculates a sphere through a set of ponts

    Args:
        vertices: entire vertex list
        indices: indices of the vertices to consider

    Returns:
        centre: centre of sphere
        radius: radius of sphere
    """
    dimension = vertices.shape[1]
    k = max(indices.shape)

    if k==1:
        # use the single value in indices as index into points
        centre = vertices[indices-1,:][0]
        radius = 0
    else:
        # split array
        a1 = vertices[indices[0]-1,:]       # first value
        a2 = vertices[indices[1:k]-1,:]     # all except first

        # reshape a1 into column vector
        a1 = a1.reshape(a1.shape[0], 1)

        aa = 2 * (a2 - (np.dot(a1, ml.ones((1, k-1)))).transpose())
        ab = np.concatenate((a1, a2.transpose()), 1)
        ac = aa.dot(ab)
        a = np.concatenate((ac, ml.ones((1, k))))

        a2d = np.diag(np.dot(a2, a2.transpose())) - np.dot(
            a1.transpose(), a1)
        b = np.concatenate((a2d[0], np.array([1])))

        # calculate the reciprocal of the conditions of a
        cn = np.reciprocal(np.linalg.cond(a))

        if cn < 1e-8:
            # return invalid values
            centre = np.empty(dimension).transpose()
            centre[:] = np.nan
            radius = np.inf
        else:
            # solve linear equation
            x, res, rank, s = np.linalg.lstsq(a, b)
            centre = np.dot(ab, x)

            atc = a1.transpose()-centre
            atcd = np.dot(atc, atc.transpose())
            radius = np.sqrt(atcd)[0][0]

    return centre, radius

def minsphere(vertices, fixed, current_basis, others, max_radius):
    """ Calculates the minimal enclosing sphere for a set of points

    Args:
        vertices: entire vertex list
        fixed: list of indices of vertices that have already been fixed
        current_basis: list of  indices of vertices that lie on the
            current sphere
        others: vertices yet to be considered
        max_radius : max radius to check for enclosing sphere

    Returns:
        centre: centre of minimal enclosing sphere
        radius: radius of minimal enclosing sphere
    """
    # find sphere through vertices contained in fixed+current_basis
    vertex_indices = np.concatenate((fixed, current_basis)).astype(int)
    c, r = through(vertices, vertex_indices)

    loop = True

    while loop:
        if others.size == 0:
            farthest = 0
        else:
            ones = np.empty(others.size)
            ones.fill(1)

            # translate the vertices so that c is the origin
            v = vertices[(others-1).astype(int), :] - np.dot(
                c.reshape(c.shape[0], 1), ones)

            # find the magnitude of each of the vertices
            distance = np.sqrt(np.sum(v**2, axis=1))

            # get the vertex with the max distance from centre of sphere
            i = np.argmax(distance)
            farthest = distance[i]

        # continue if farthest is outside the current basis
        loop = (farthest>r+mutil.epsilon) and (r<max_radius-mutil.epsilon)

        if loop:
            # get the furthest vector
            out = others[i]

            # recursive call to minsphere with furthest as fixed
            newfixed = np.concatenate((fixed, np.array([out])))
            c, r, newbasis = minsphere(vertices, newfixed, np.array([]),
                current_basis, max_radius)

            # get the others except for the farthest
            others_a = np.concatenate((others[0:i], others[(i+1):]))

            # get the values in current_basis not in newbasis
            basis_diff = np.setdiff1d(current_basis, newbasis)

            # set the new others
            others = np.concatenate((others_a, basis_diff))
            current_basis = np.concatenate((newbasis, np.array([out])))

    return c, r, current_basis
