import numpy as np
import numpy.linalg as la
import scipy.optimize as opt

def simplical_complex_entropy(a, p):
    """ Calculates an entropy function for simplical complices

    Args:
        a: incidence matrix
        p: probability distribution

    Returns:
        e: entropy
        err: error
        q: value of objective function
    """
    m, n = a.shape

    def calculate_entropy(q):
        qda = np.dot(q.transpose(), a)
        qda_log = np.log2(qda)

        y = np.dot(-qda_log, p)

        ones = np.empty(n)
        ones.fill(1)

        adq = np.dot(a.transpose(), q)
        b = np.divide(ones, adq)

        dy = np.dot(np.dot(-a, np.diag(p)), b)

        return y, dy

    q0 = np.random.rand(m, 1)
    q0 = np.divide(q0, la.norm(q0, 1))

    # create the bounds for each element in x (0, 1)
    bounds = []
    for i in range(0, m):
        bounds.append((0, 1))

    opt_result = opt.minimize(calculate_entropy, q0, method='SLSQP',
        jac=True, bounds=bounds, tol=1e-6, options={'ftol': 1e-6,
        'maxiter': 100000})

    # get the results and objective function value
    q = np.absolute(opt_result.jac[:opt_result.jac.shape[0]-1])
    e = opt_result.fun

    b = np.dot(a.transpose(), q)

    sum_pda = np.sum(np.dot(np.diag(p), a.transpose()), axis=0).transpose()

    ones = np.empty(m)
    ones.fill(1)

    max_pda = np.amax(np.dot(np.diag(p), a.transpose()), axis=0)

    c = ones - np.divide(max_pda, sum_pda)

    ap = a.transpose()
    lin_result = opt.linprog(c, np.empty(ap.shape), np.empty(ap.shape[0]),
        ap, b, (0, 1), options={'tol': 1e-06})

    q = lin_result.x
    err = lin_result.fun

    return e, err, q
